
/* Chrysalide - Outil d'analyse de fichiers binaires
 * mclf.c - prise en compte du format binaire 'MCLF'
 *
 * Copyright (C) 2015-2017 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "mclf.h"


#include <i18n.h>


#include "mclf-int.h"
#include "symbols.h"



/* Taille maximale d'une description */
#define MAX_PORTION_DESC 256



/* Initialise la classe des formats d'exécutables MCLF. */
static void g_mclf_format_class_init(GMCLFFormatClass *);

/* Initialise une instance de format d'exécutable MCLF. */
static void g_mclf_format_init(GMCLFFormat *);

/* Supprime toutes les références externes. */
static void g_mclf_format_dispose(GMCLFFormat *);

/* Procède à la libération totale de la mémoire. */
static void g_mclf_format_finalize(GMCLFFormat *);

/* Indique le type d'architecture visée par le format. */
static const char *g_mclf_format_get_target_machine(const GMCLFFormat *);

/* Etend la définition des portions au sein d'un binaire. */
static void g_mclf_format_refine_portions(GMCLFFormat *);



/******************************************************************************
*                                                                             *
*  Paramètres  : content = contenu binaire à parcourir.                       *
*                parent  = éventuel format exécutable déjà chargé.            *
*                unused  = adresse non utilisée ici.                          *
*                key     = identifiant de format trouvé ou NULL. [OUT]        *
*                                                                             *
*  Description : Indique si le format peut être pris en charge ici.           *
*                                                                             *
*  Retour      : Désignation du format reconnu ou NULL si aucun.              *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

FormatMatchStatus mclf_is_matching(GBinContent *content, GExeFormat *parent, void *unused, char **key)
{
    FormatMatchStatus result;               /* Bilan à renvoyer            */
    vmpa2t addr;                            /* Tête de lecture initiale    */
    bool status;                            /* Bilan des accès mémoire     */
    char magic[4];                          /* Idenfiant standard          */

    init_vmpa(&addr, 0, VMPA_NO_VIRTUAL);

    status = g_binary_content_read_raw(content, &addr, 4, (bin_t *)magic);

    status &= (memcmp(magic, MC_SERVICE_HEADER_MAGIC_STR, 4) == 0);

    if (status)
    {
        result = FMS_MATCHED;
        *key = strdup("mclf");
    }
    else
        result = FMS_UNKNOWN;

    return result;

}


/* Indique le type défini pour un format d'exécutable MCLF. */
G_DEFINE_TYPE(GMCLFFormat, g_mclf_format, G_TYPE_EXE_FORMAT);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des formats d'exécutables MCLF.         *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_mclf_format_class_init(GMCLFFormatClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GExeFormatClass *exe;                   /* Version en exécutable       */

    object = G_OBJECT_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_mclf_format_dispose;
    object->finalize = (GObjectFinalizeFunc)g_mclf_format_finalize;

    exe = G_EXE_FORMAT_CLASS(klass);

    exe->get_machine = (get_target_machine_fc)g_mclf_format_get_target_machine;
    exe->refine_portions = (refine_portions_fc)g_mclf_format_refine_portions;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : format = instance à initialiser.                             *
*                                                                             *
*  Description : Initialise une instance de format d'exécutable MCLF.         *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_mclf_format_init(GMCLFFormat *format)
{

}


/******************************************************************************
*                                                                             *
*  Paramètres  : format = instance d'objet GLib à traiter.                    *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_mclf_format_dispose(GMCLFFormat *format)
{
    G_OBJECT_CLASS(g_mclf_format_parent_class)->dispose(G_OBJECT(format));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : format = instance d'objet GLib à traiter.                    *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_mclf_format_finalize(GMCLFFormat *format)
{
    G_OBJECT_CLASS(g_mclf_format_parent_class)->finalize(G_OBJECT(format));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : content = contenu binaire à parcourir.                       *
*                parent  = éventuel format exécutable déjà chargé.            *
                 status  = barre de statut à tenir informée.                  *
*                                                                             *
*  Description : Prend en charge un nouveau format MCLF.                      *
*                                                                             *
*  Retour      : Adresse de la structure mise en place ou NULL en cas d'échec.*
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GBinFormat *g_mclf_format_new(GBinContent *content, GExeFormat *parent, GtkStatusStack *status)
{
    GMCLFFormat *result;                    /* Structure à retourner       */

    result = g_object_new(G_TYPE_MCLF_FORMAT, NULL);

    g_binary_format_set_content(G_BIN_FORMAT(result), content);



    if (!read_mclf_header(result, &result->header, result->endian))
    {
        /* TODO */
        return NULL;
    }



    if (!load_mclf_symbols(result))
    {
        /* TODO */
        return NULL;
    }


    return G_BIN_FORMAT(result);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : format = informations chargées à consulter.                  *
*                                                                             *
*  Description : Indique le type d'architecture visée par le format.          *
*                                                                             *
*  Retour      : Identifiant de l'architecture ciblée par le format.          *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static const char *g_mclf_format_get_target_machine(const GMCLFFormat *format)
{
    const char *result;                     /* Identifiant à retourner     */

    result = "armv7";

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : format = informations chargées à consulter.                  *
*                                                                             *
*  Description : Etend la définition des portions au sein d'un binaire.       *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_mclf_format_refine_portions(GMCLFFormat *format)
{
    GExeFormat *exe_format;                 /* Autre version du format     */
    vmpa2t addr;                            /* Emplacement dans le binaire */
    GBinPortion *new;                       /* Nouvelle portion définie    */
    char desc[MAX_PORTION_DESC];            /* Description d'une portion   */
    phys_t length;                          /* Taille de portion globale   */

    exe_format = G_EXE_FORMAT(format);

    /* Segment de code */

    init_vmpa(&addr, 0, format->header.v1.text.start);

    new = g_binary_portion_new(BPC_CODE, &addr, format->header.v1.text.len);

    sprintf(desc, "%s \"%s\"", _("Segment"), "text");
    g_binary_portion_set_desc(new, desc);

    g_binary_portion_set_rights(new, PAC_WRITE | PAC_EXEC);

    g_exe_format_include_portion(exe_format, new);

    /* Segment de données */

    init_vmpa(&addr, format->header.v1.text.len, format->header.v1.data.start);

    new = g_binary_portion_new(BPC_DATA, &addr, format->header.v1.data.len);

    sprintf(desc, "%s \"%s\"", _("Segment"), "data");
    g_binary_portion_set_desc(new, desc);

    g_binary_portion_set_rights(new, PAC_READ | PAC_WRITE);

    g_exe_format_include_portion(exe_format, new);

    /* Signature finale */

    length = g_binary_content_compute_size(G_BIN_FORMAT(format)->content);
    init_vmpa(&addr, length - 521, VMPA_NO_VIRTUAL);  /* FIXME */

    new = g_binary_portion_new(BPC_DATA, &addr, 521);

    sprintf(desc, "%s \"%s\"", _("Segment"), "sig");
    g_binary_portion_set_desc(new, desc);

    g_binary_portion_set_rights(new, PAC_READ | PAC_WRITE);

    g_exe_format_include_portion(exe_format, new);

}
