
/* Chrysalide - Outil d'analyse de fichiers binaires
 * args.c - listes d'opérandes rassemblées en arguments
 *
 * Copyright (C) 2010-2017 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "args.h"


#include <assert.h>
#include <malloc.h>


#include "../../operand-int.h"
#include "../../sharing/manager.h"
#include "../../../common/sort.h"



/* --------------------- MANIPULATION D'OPERANDES INDIVIDUELLES --------------------- */


/* Définition d'un opérande visant une liste d'opérandes Dalvik (instance) */
struct _GDalvikArgsOperand
{
    GArchOperand parent;                    /* Instance parente            */

    GArchOperand **args;                    /* Liste d'arguments           */
    size_t count;                           /* Taille de cette liste       */

};


/* Définition d'un opérande visant une liste d'opérandes Dalvik (classe) */
struct _GDalvikArgsOperandClass
{
    GArchOperandClass parent;               /* Classe parente              */

};


/* Initialise la classe des listes d'opérandes Dalvik. */
static void g_dalvik_args_operand_class_init(GDalvikArgsOperandClass *);

/* Initialise une instance de liste d'opérandes Dalvik. */
static void g_dalvik_args_operand_init(GDalvikArgsOperand *);

/* Supprime toutes les références externes. */
static void g_dalvik_args_operand_dispose(GDalvikArgsOperand *);

/* Procède à la libération totale de la mémoire. */
static void g_dalvik_args_operand_finalize(GDalvikArgsOperand *);

/* Initialise un nouvel objet partagé avec des informations. */
static bool g_dalvik_args_operand_apply_template(GDalvikArgsOperand *, const GDalvikArgsOperand *);

/* Réalise une copie minimale d'un contenu partagé. */
static void g_dalvik_args_operand_define_template(const GDalvikArgsOperand *, GDalvikArgsOperand *);

/* Libère la mémoire utilisée par un patron d'instance. */
static void g_dalvik_args_operand_free_template(const GDalvikArgsOperand *, GDalvikArgsOperand *);

/* Compare un opérande avec un autre. */
static int g_dalvik_args_operand_compare(const GDalvikArgsOperand *, const GDalvikArgsOperand *);

/* Traduit un opérande en version humainement lisible. */
static void g_dalvik_args_operand_print(const GDalvikArgsOperand *, GBufferLine *, AsmSyntax);



/* -------------------------- PARTAGES DE CONTENUS UNIQUES -------------------------- */


/* Gestionnaire des partages d'instances */
static GShareManager *_dalvik_args_operand_manager = NULL;


/* Fournit le gestionnaire de partages attribué à un type. */
static GShareManager *get_dalvik_args_operand_share_manager(void);



/* ---------------------------------------------------------------------------------- */
/*                       MANIPULATION D'OPERANDES INDIVIDUELLES                       */
/* ---------------------------------------------------------------------------------- */


/* Indique le type défini par la GLib pour une liste d'arguments Dalvik. */
G_DEFINE_TYPE(GDalvikArgsOperand, g_dalvik_args_operand, G_TYPE_ARCH_OPERAND);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des listes d'opérandes Dalvik.          *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_dalvik_args_operand_class_init(GDalvikArgsOperandClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GArchOperandClass *operand;             /* Version de classe parente   */

    object = G_OBJECT_CLASS(klass);
    operand = G_ARCH_OPERAND_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_dalvik_args_operand_dispose;
    object->finalize = (GObjectFinalizeFunc)g_dalvik_args_operand_finalize;

    operand->get_manager = (get_operand_manager_fc)get_dalvik_args_operand_share_manager;

    operand->apply_template = (apply_operand_template_fc)g_dalvik_args_operand_apply_template;
    operand->define_template = (define_operand_template_fc)g_dalvik_args_operand_define_template;
    operand->free_template = (free_operand_template_fc)g_dalvik_args_operand_free_template;

    operand->compare = (operand_compare_fc)g_dalvik_args_operand_compare;
    operand->print = (operand_print_fc)g_dalvik_args_operand_print;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = instance à initialiser.                            *
*                                                                             *
*  Description : Initialise une instance de liste d'opérandes Dalvik.         *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_dalvik_args_operand_init(GDalvikArgsOperand *operand)
{

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = instance d'objet GLib à traiter.                   *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_dalvik_args_operand_dispose(GDalvikArgsOperand *operand)
{
    size_t i;

    for (i = 0; i < operand->count; i++)
        g_object_unref(G_OBJECT(operand->args[i]));

    G_OBJECT_CLASS(g_dalvik_args_operand_parent_class)->dispose(G_OBJECT(operand));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = instance d'objet GLib à traiter.                   *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_dalvik_args_operand_finalize(GDalvikArgsOperand *operand)
{
    G_OBJECT_CLASS(g_dalvik_args_operand_parent_class)->finalize(G_OBJECT(operand));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand  = objet partagé à initialiser.                      *
*                template = information à utiliser pour la mise en place.     *
*                                                                             *
*  Description : Initialise un nouvel objet partagé avec des informations.    *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_dalvik_args_operand_apply_template(GDalvikArgsOperand *operand, const GDalvikArgsOperand *template)
{
    size_t i;                               /* Boucle de parcours          */

    if (template == NULL)
    {
        operand->args = NULL;
        operand->count = 0;
    }

    else
    {
        g_dalvik_args_operand_define_template(template, operand);

        for (i = 0; i < operand->count; i++)
            g_object_ref(G_OBJECT(operand->args[i]));

    }

    return true;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand  = objet partagé à consulter.                        *
*                template = informations à retrouver intégralement.           *
*                                                                             *
*  Description : Réalise une copie minimale d'un contenu partagé.             *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_dalvik_args_operand_define_template(const GDalvikArgsOperand *operand, GDalvikArgsOperand *template)
{
    size_t i;                               /* Boucle de parcours          */

    if (operand->count == 0)
    {
        template->args = NULL;
        template->count = 0;
    }

    else
    {
        template->args = (GArchOperand **)calloc(operand->count, sizeof(GArchOperand *));

        for (i = 0; i < operand->count; i++)
            template->args[i] = operand->args[i];

        template->count = operand->count;

    }

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand  = objet partagé à consulter.                        *
*                template = informations dont le contenu est à libérer.       *
*                                                                             *
*  Description : Libère la mémoire utilisée par un patron d'instance.         *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_dalvik_args_operand_free_template(const GDalvikArgsOperand *operand, GDalvikArgsOperand *template)
{
    if (template != NULL && template->args != NULL)
        free(template->args);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : a = premier opérande à consulter.                            *
*                b = second opérande à consulter.                             *
*                                                                             *
*  Description : Compare un opérande avec un autre.                           *
*                                                                             *
*  Retour      : Bilan de la comparaison.                                     *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static int g_dalvik_args_operand_compare(const GDalvikArgsOperand *a, const GDalvikArgsOperand *b)
{
    int result;                             /* Bilan à renvoyer            */
    size_t i;                               /* Boucle de parcours          */

    /* Création de l'objet... */
    if (b == NULL)
        result = 1;

    else
    {
        result = sort_unsigned_long(a->count, b->count);

        for (i = 0; i < a->count && result == 0; i++)
            result = g_arch_operand_compare(a->args[i], b->args[i]);

    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = opérande à traiter.                                *
*                line    = ligne tampon où imprimer l'opérande donné.         *
*                syntax  = type de représentation demandée.                   *
*                                                                             *
*  Description : Traduit un opérande en version humainement lisible.          *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_dalvik_args_operand_print(const GDalvikArgsOperand *operand, GBufferLine *line, AsmSyntax syntax)
{
    size_t i;                               /* Boucle de parcours          */

    g_buffer_line_append_text(line, BLC_ASSEMBLY, "{", 1, RTT_HOOK, NULL);

    if (operand->count > 0)
    {
        g_arch_operand_print(operand->args[0], line, syntax);

        for (i = 1; i < operand->count; i++)
        {
            g_buffer_line_append_text(line, BLC_ASSEMBLY, ",", 1, RTT_PUNCT, NULL);
            g_buffer_line_append_text(line, BLC_ASSEMBLY, " ", 1, RTT_RAW, NULL);

            g_arch_operand_print(operand->args[i], line, syntax);

        }

    }

    g_buffer_line_append_text(line, BLC_ASSEMBLY, "}", 1, RTT_HOOK, NULL);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Crée un réceptacle pour opérandes Dalvik servant d'arguments.*
*                                                                             *
*  Retour      : Opérande mis en place.                                       *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GArchOperand *g_dalvik_args_operand_new(void)
{
    GArchOperand *result;                   /* Structure à retourner       */

    result = G_ARCH_OPERAND(g_share_manager_build(_dalvik_args_operand_manager, NULL));

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = opérande à compléter.                              *
*                arg     = nouvel argument pour un appel.                     *
*                                                                             *
*  Description : Ajoute un élément à la liste d'arguments Dalvik.             *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GDalvikArgsOperand *g_dalvik_args_operand_add(GDalvikArgsOperand *operand, GArchOperand *arg, GShareContainer *container)
{
    GSharedInstance *result;                /* Nouvelle version à renvoyer */
    GDalvikArgsOperand fake;                /* Transport d'informations    */
    size_t i;                               /* Boucle de parcours          */

    fake.count = operand->count + 1;
    fake.args = (GArchOperand **)calloc(fake.count, sizeof(GArchOperand *));

    for (i = 0; i < operand->count; i++)
        fake.args[i] = operand->args[i];

    fake.args[i] = arg;

    result = g_share_manager_update(_dalvik_args_operand_manager, G_SHARED_INSTANCE(operand),
                                    (GSharedInstance *)&fake, container);

    free(fake.args);

    return G_DALVIK_ARGS_OPERAND(result);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = opérande à compléter.                              *
*                                                                             *
*  Description : Fournit le nombre d'arguments pris en charge.                *
*                                                                             *
*  Retour      : Nombre positif ou nul.                                       *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

size_t g_dalvik_args_count(const GDalvikArgsOperand *operand)
{
    return operand->count;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = opérande à compléter.                              *
*                index   = indice de l'argument recherché.                    *
*                                                                             *
*  Description : Founit un élément de la liste d'arguments Dalvik.            *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GArchOperand *g_dalvik_args_operand_get(const GDalvikArgsOperand *operand, size_t index)
{
    assert(index < operand->count);

    return operand->args[index];

}



/* ---------------------------------------------------------------------------------- */
/*                            PARTAGES DE CONTENUS UNIQUES                            */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Fournit le gestionnaire de partages attribué à un type.      *
*                                                                             *
*  Retour      : Gestionnaire de partages en place.                           *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static GShareManager *get_dalvik_args_operand_share_manager(void)
{
    return _dalvik_args_operand_manager;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Met en place les mécanismes de partage des opérandes Dalvik. *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool init_dalvik_args_operand_sharing(void)
{
    _dalvik_args_operand_manager = g_share_manager_new(G_TYPE_DALVIK_ARGS_OPERAND);

    return true;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Imprime des statistiques quant aux partages dans l'archi.    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/
#ifdef DEBUG_DUMP_STATS
void dump_dalvik_args_operand_share_stats(void)
{
    g_share_manager_dump_stats(_dalvik_args_operand_manager);

}
#endif


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Supprime les mécanismes de partage des opérandes Dalvik.     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void exit_dalvik_args_operand_sharing(void)
{
    g_object_unref(G_OBJECT(_dalvik_args_operand_manager));

}
